@@ -187,7 +187,7 @@ impl CachePolicy {
         res: &Res,
         opts: CachePolicyOptions,
     ) -> Self {
-        let uri = req.uri().clone();
+        let uri = req.uri();
         let status = res.status();
         let method = req.method().clone();
         let res = res.headers().clone();
@@ -360,7 +360,7 @@ impl CachePolicy {
 
     fn request_matches<Req: RequestLike>(&self, req: &Req, allow_head_method: bool) -> bool {
         // The presented effective request URI and that of the stored response match, and
-        &self.uri == req.uri() &&
+        req.is_same_uri(&self.uri) &&
             self.req.get("host") == req.headers().get("host") &&
             // the request method associated with the stored response allows it to be used for the presented request, and
             (
@@ -423,7 +423,10 @@ impl CachePolicy {
     /// returning it to a client. This function is necessary, because proxies
     /// MUST always remove hop-by-hop headers (such as TE and Connection) and
     /// update response's Age to avoid doubling cache time.
-    pub fn response_headers(&self, now: SystemTime) -> HeaderMap {
+    ///
+    /// It returns response "parts" without a body. You can upgrade it to a full
+    /// response with `Response::from_parts(parts, BYOB)`
+    pub fn cached_response(&self, now: SystemTime) -> http::response::Parts {
         let mut headers = Self::copy_without_hop_by_hop_headers(&self.res);
         let age = self.age(now);
         let day = Duration::from_secs(3600 * 24);
@@ -446,7 +449,14 @@ impl CachePolicy {
             HeaderValue::from_str(&format!("{}", age.as_secs() as u32)).unwrap(),
         );
         headers.insert("date", HeaderValue::from_str(&date.to_rfc2822()).unwrap());
-        headers
+
+        let mut parts = Response::builder()
+            .status(self.status)
+            .body(())
+            .unwrap()
+            .into_parts().0;
+        parts.headers = headers;
+        parts
     }
 
     /// Value of the Date response header or current time if Date was demed invalid
@@ -581,7 +591,7 @@ impl CachePolicy {
     /// After that time (when `time_to_live() <= 0`) the response might not be
     /// usable without revalidation. However, there are exceptions, e.g. a
     /// client can explicitly allow stale responses, so always check with
-    /// `is_cached_response_fresh()`.
+    /// `satisfies_without_revalidation()`.
     pub fn time_to_live(&self, now: SystemTime) -> Duration {
         self.max_age()
             .checked_sub(self.age(now))
@@ -599,7 +609,9 @@ impl CachePolicy {
     /// Hop by hop headers are always stripped.
     /// Revalidation headers may be added or removed, depending on request.
     ///
-    pub fn revalidation_headers<Req: RequestLike>(&self, incoming_req: &Req) -> HeaderMap {
+    /// It returns request "parts" without a body. You can upgrade it to a full
+    /// response with `Request::from_parts(parts, BYOB)` (the body is usually `()`).
+    pub fn revalidation_request<Req: RequestLike>(&self, incoming_req: &Req) -> http::request::Parts {
         let mut headers = Self::copy_without_hop_by_hop_headers(incoming_req.headers());
 
         // This implementation does not understand range requests
@@ -610,7 +622,7 @@ impl CachePolicy {
             // not for the same resource, or wasn't allowed to be cached anyway
             headers.remove("if-none-match");
             headers.remove("if-modified-since");
-            return headers;
+            return self.request_from_headers(headers);
         }
 
         /* MUST send that entity-tag in any cache validation request (using If-Match or If-None-Match) if an entity-tag has been provided by the origin server. */
@@ -647,8 +659,18 @@ impl CachePolicy {
                 );
             }
         }
+        self.request_from_headers(headers)
+    }
 
-        headers
+    fn request_from_headers(&self, headers: HeaderMap) -> http::request::Parts {
+        let mut parts = Request::builder()
+            .method(self.method.clone())
+            .uri(self.uri.clone())
+            .body(())
+            .unwrap()
+            .into_parts().0;
+        parts.headers = headers;
+        parts
     }
 
     /// Creates `CachePolicy` with information combined from the previews response,
@@ -656,15 +678,19 @@ impl CachePolicy {
     ///
     /// Returns `{policy, modified}` where modified is a boolean indicating
     /// whether the response body has been modified, and old cached body can't be used.
-    pub fn revalidated_policy<ReqB, ResB>(
+    pub fn revalidated_policy<Req: RequestLike, Res: ResponseLike>(
         &self,
-        request: &Request<ReqB>,
-        response: &mut Response<ResB>,
+        request: &Req,
+        response: &Res,
+        response_time: SystemTime
     ) -> RevalidatedPolicy {
-        let old_etag = self.res.get_str("etag").map(str::trim);
-        let old_last_modified = response.headers().get_str("last-modified").map(str::trim);
-        let new_etag = response.headers().get_str("new_etag").map(str::trim);
-        let new_last_modified = response.headers().get_str("last-modified").map(str::trim);
+        let response_headers = response.headers();
+        let mut response_status = response.status();
+
+        let old_etag = &self.res.get_str("etag").map(str::trim);
+        let old_last_modified = response_headers.get_str("last-modified").map(str::trim);
+        let new_etag = response_headers.get_str("etag").map(str::trim);
+        let new_last_modified = response_headers.get_str("last-modified").map(str::trim);
 
         // These aren't going to be supported exactly, since one CachePolicy object
         // doesn't know about all the other cached objects.
@@ -698,19 +724,33 @@ impl CachePolicy {
         }
 
         let modified = response.status() != StatusCode::NOT_MODIFIED;
-        if matches {
+        let new_response_headers = if matches {
+            let mut new_response_headers = HeaderMap::with_capacity(self.res.keys_len());
             // use other header fields provided in the 304 (Not Modified) response to replace all instances
             // of the corresponding header fields in the stored response.
-            for (h, v) in &self.res {
-                if !EXCLUDED_FROM_REVALIDATION_UPDATE.contains(&&h.as_str()) {
-                    response.headers_mut().insert(h.to_owned(), v.to_owned());
+            for (header, old_value) in &self.res {
+                let header = header.to_owned();
+                if let Some(new_value) = response_headers.get(&header) {
+                    if !EXCLUDED_FROM_REVALIDATION_UPDATE.contains(&&header.as_str()) {
+                        new_response_headers.insert(header, new_value.to_owned());
+                        continue;
+                    }
                 }
+                new_response_headers.insert(header, old_value.to_owned());
             }
-            *response.status_mut() = self.status;
-        }
+            response_status = self.status;
+            new_response_headers
+        } else {
+            response_headers.clone()
+        };
+
+        let mut new_response = Response::builder().status(response_status).body(()).unwrap().into_parts().0;
+        new_response.headers = new_response_headers;
+        let mut new_options = self.opts;
+        new_options.response_time = response_time;
 
         RevalidatedPolicy {
-            policy: CachePolicy::new(request, &*response, self.opts),
+            policy: CachePolicy::new(request, &new_response, new_options),
             // Client receiving 304 without body, even if it's invalid/mismatched has no option
             // but to reuse a cached body. We don't have a good way to tell clients to do
             // error recovery in such case.
@@ -765,8 +805,12 @@ fn join<'a>(parts: impl Iterator<Item = &'a str>) -> String {
 
 /// Allows using either Request or request::Parts, or your own newtype.
 pub trait RequestLike {
-    /// Same as `req.uri()`
-    fn uri(&self) -> &Uri;
+    /// Same as `req.uri().clone()`
+    fn uri(&self) -> Uri;
+    /// Whether the effective request URI matches the other URI
+    ///
+    /// It can be naive string comparison, nothing fancy
+    fn is_same_uri(&self, other: &Uri) -> bool;
     /// Same as `req.method()`
     fn method(&self) -> &Method;
     /// Same as `req.headers()`
@@ -782,8 +826,11 @@ pub trait ResponseLike {
 }
 
 impl<Body> RequestLike for Request<Body> {
-    fn uri(&self) -> &Uri {
-        self.uri()
+    fn uri(&self) -> Uri {
+        self.uri().clone()
+    }
+    fn is_same_uri(&self, other: &Uri) -> bool {
+        self.uri() == other
     }
     fn method(&self) -> &Method {
         self.method()
@@ -794,8 +841,11 @@ impl<Body> RequestLike for Request<Body> {
 }
 
 impl RequestLike for http::request::Parts {
-    fn uri(&self) -> &Uri {
-        &self.uri
+    fn uri(&self) -> Uri {
+        self.uri.clone()
+    }
+    fn is_same_uri(&self, other: &Uri) -> bool {
+        &self.uri == other
     }
     fn method(&self) -> &Method {
         &self.method
@@ -822,3 +872,29 @@ impl ResponseLike for http::response::Parts {
         &self.headers
     }
 }
+
+#[cfg(feature = "reqwest")]
+impl RequestLike for reqwest::Request {
+    fn uri(&self) -> Uri {
+        self.url().as_str().parse().expect("Uri and Url are incompatible!?")
+    }
+    fn is_same_uri(&self, other: &Uri) -> bool {
+        self.url().as_str() == other
+    }
+    fn method(&self) -> &Method {
+        self.method()
+    }
+    fn headers(&self) -> &HeaderMap {
+        self.headers()
+    }
+}
+
+#[cfg(feature = "reqwest")]
+impl ResponseLike for reqwest::Response {
+    fn status(&self) -> StatusCode {
+        self.status()
+    }
+    fn headers(&self) -> &HeaderMap {
+        self.headers()
+    }
+}
